import numpy as np


def volts_to_bytes(volt=None, dac_bitMax=None):
    return np.ceil(((volt + 10) / float(20)) * dac_bitMax)


def encode_message(*message_parts, encoding=None):
    message = [np.array(part, dtype=encoding).tobytes() for part in message_parts]
    out = b"".join(message)
    return out



# https://sites.google.com/site/pulsepalwiki/matlab-gnu-octave/functions/programpulsepalparam
PARAM_CODES = {
    1: "isBiphasic",
    2: "phase1Voltage",
    3: "phase2Voltage",
    4: "phase1Duration",
    5: "interPhaseInterval",
    6: "phase2Duration",
    7: "interPulseInterval",
    8: "burstDuration",
    9: "interBurstInterval",
    10: "pulseTrainDuration",
    11: "pulseTrainDelay",
    12: "linkTriggerChannel1",
    13: "linkTriggerChannel2",
    14: "customTrainID",
    15: "customTrainTarget",
    16: "customTrainLoop",
    17: "restingVoltage",
    128: "triggerMode",
}
PARAM_NAMES = {v: k for k, v in PARAM_CODES.items()}
CHANNEL_PARAM_DEFAULTS = {
    "isBiphasic": False,
    "phase1Voltage": 5,
    "phase2Voltage": -5,
    "phase1Duration": 0.001,
    "interPhaseInterval": 0.001,
    "phase2Duration": 0.001,
    "interPulseInterval": 0.01,
    "burstDuration": 0,
    "interBurstInterval": 0,
    "pulseTrainDuration": 1,
    "pulseTrainDelay": 0,
    "linkTriggerChannel1": 1,
    "linkTriggerChannel2": 0,
    "customTrainID": 0,
    "customTrainTarget": 0,
    "customTrainLoop": 0,
    "restingVoltage": 0,
    # "triggerMode": 0,  # not a channel param
}
CHANNEL_PARAM_TEST = {
    "isBiphasic": False,
    "phase1Voltage": 1.85,
    "phase2Voltage": -2.45,
    "phase1Duration": 0.029,
    "interPhaseInterval": 0.251,
    "phase2Duration": 0.051,
    "interPulseInterval": 0.111,
    "burstDuration": 1.1,
    "interBurstInterval": 2.66,
    "pulseTrainDuration": 4.55,
    "pulseTrainDelay": 0.444,
    "linkTriggerChannel1": 0,
    "linkTriggerChannel2": 1,
    "customTrainID": 0,
    "customTrainTarget": 0,
    "customTrainLoop": 0,
    "restingVoltage": 4.99,
    "triggerMode": 1,  # not a channel param
}
TRIGGER_PARAM_TEST = {
    "triggerMode": 2,
}
TRIGGER_PARAM_DEFAULTS = {
    "triggerMode": 0,
}
TRIGGER_MODE_VALUES = {
    0: "normal",
    1: "toggle",
    2: "gated",
}
TRIGGER_MODE_NAMES = {v: k for k, v in TRIGGER_MODE_VALUES.items()}
# TYPES:
#   volt @ [2,3,17] -> "uint16",
#   ELIF [4-11] -> "uint32",
#   ELSE (for 1/biphasic, [12-16]/custom*, 128/triggerMode) -> "uint8"
PARAM_DTYPE_MODEL_1 = {
    "isBiphasic": "uint8",
    "phase1Voltage": "uint8",  # v1: uint8
    "phase2Voltage": "uint8",  # v1: uint8
    "phase1Duration": "uint32",  # param 4
    "interPhaseInterval": "uint32",
    "phase2Duration": "uint32",
    "interPulseInterval": "uint32",
    "burstDuration": "uint32",
    "interBurstInterval": "uint32",
    "pulseTrainDuration": "uint32",
    "pulseTrainDelay": "uint32",
    "linkTriggerChannel1": "uint8",
    "linkTriggerChannel2": "uint8",
    "customTrainID": "uint8",
    "customTrainTarget": "uint8",
    "customTrainLoop": "uint8",
    "restingVoltage": "uint8",  # v1: uint8, param: 17
    "triggerMode": "uint8",  # parma 128
}
PARAM_DTYPE_MODEL_2 = {
    "isBiphasic": "uint8",
    "phase1Voltage": "uint16",  # v1: uint8
    "phase2Voltage": "uint16",
    "phase1Duration": "uint32",  # param 4
    "interPhaseInterval": "uint32",
    "phase2Duration": "uint32",
    "interPulseInterval": "uint32",
    "burstDuration": "uint32",
    "interBurstInterval": "uint32",
    "pulseTrainDuration": "uint32",
    "pulseTrainDelay": "uint32",
    "linkTriggerChannel1": "uint8",
    "linkTriggerChannel2": "uint8",
    "customTrainID": "uint8",
    "customTrainTarget": "uint8",
    "customTrainLoop": "uint8",
    "restingVoltage": "uint16",  # v1: uint8, param: 17
    "triggerMode": "uint8",  # parma 128
}
PULSEPAL_CYCLE_FREQUENCY = 20000
PARAM_SCALING = {
    "isBiphasic": 1,
    "phase1Voltage": 1,
    "phase2Voltage": 1,
    "phase1Duration": PULSEPAL_CYCLE_FREQUENCY,  # param 4
    "interPhaseInterval": PULSEPAL_CYCLE_FREQUENCY,
    "phase2Duration": PULSEPAL_CYCLE_FREQUENCY,
    "interPulseInterval": PULSEPAL_CYCLE_FREQUENCY,
    "burstDuration": PULSEPAL_CYCLE_FREQUENCY,
    "interBurstInterval": PULSEPAL_CYCLE_FREQUENCY,
    "pulseTrainDuration": PULSEPAL_CYCLE_FREQUENCY,
    "pulseTrainDelay": PULSEPAL_CYCLE_FREQUENCY,  # param 11
    "linkTriggerChannel1": 1,
    "linkTriggerChannel2": 1,
    "customTrainID": 1,
    "customTrainTarget": 1,
    "customTrainLoop": 1,
    "restingVoltage": 1,  # param: 17
    "triggerMode": 1,  # parma 128
}


class SendMessageHeader:
    """
    https://sites.google.com/site/pulsepalwiki/usb-serial-interface/usb-interface-v2-x
    """

    HANDSHAKE = "H"  # "H" == chr(72)

    PROGRAM_ALL = 73
    PROGRAM_ONE = 74  # next: param code, then channel, then time/voltage
    PROGRAM_CUSTOM_1 = 75
    PROGRAM_CUSTOM_2 = 76
    SOFT_TRIGGER = 77
    DISPLAY = 78
    PROGRAM_VOLT = 79

    ABORT_ALL = 80
    DISCONNECT = 81

    CONTINUOUS = 82  # channel, state

    CLIENT_ID = 89  # 6 char, e.g. python/matlab

    SETTINGS = 90

    LOGIC_SET = 86
    LOGIC_GET = 87


class ReceiveMessageHeader:
    """
    https://sites.google.com/site/pulsepalwiki/usb-serial-interface/usb-interface-v2-x
    """

    HANDSHAKE_OK = "K"  # "K" == chr(75)  # + 32-bit int for firmware version

    PROGRAM_ALL_OK = 1
    PROGRAM_ONE_OK = 1


CUSTOM_PULSE_TRAIN_OPCODES = {
    0: SendMessageHeader.PROGRAM_CUSTOM_1,
    1: SendMessageHeader.PROGRAM_CUSTOM_1,
}


def resolve_param_name_code_pair(param_name_or_code=None):
    """Expect parameter name or code (integer) and return both name and code"""
    param_name = param_code = param_name_or_code
    if isinstance(param_name_or_code, str):
        param_code = PARAM_NAMES.get(param_name_or_code, None)
    elif isinstance(param_name_or_code, int):
        param_name = PARAM_CODES.get(param_name_or_code, None)
    else:
        raise ValueError(param_name_or_code)

    assert param_name is not None and param_code is not None
    return param_name, param_code


def resolve_trigger_name_code_pair(trigger_name_or_code=None):
    trigger_name = trigger_code = trigger_name_or_code
    if isinstance(trigger_name_or_code, str):
        trigger_code = PARAM_NAMES.get(trigger_name_or_code, None)
    elif isinstance(trigger_name_or_code, int):
        trigger_name = PARAM_CODES.get(trigger_name_or_code, None)
    else:
        raise ValueError(trigger_name_or_code)

    assert trigger_name is not None and trigger_code is not None
    return trigger_name, trigger_code



from pypulsepal import PulsePal
from pypulsepal.definitions import CHANNEL_PARAM_TEST
from pypulsepal.definitions import TRIGGER_PARAM_TEST


def write_test_settings_for_manual_check(serial_port=None):
    pulsepal_object = PulsePal(serial_port=serial_port)

    print("ENTER test write")
    for k, v in CHANNEL_PARAM_TEST.items():
        setattr(pulsepal_object, k, [v] * pulsepal_object.nr_output_channels)

    for k, v in TRIGGER_PARAM_TEST.items():
        setattr(pulsepal_object, k, [v] * pulsepal_object.nr_trigger_channels)

    pulsepal_object.upload_all()
    print("EXIT test write")


def write_default_settings(serial_port=None):
    print("ENTER default write")
    pulsepal_object = PulsePal(serial_port=serial_port)
    pulsepal_object.upload_all()
    print("EXIT default write")



from os import path

from setuptools import find_packages
from setuptools import setup


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()

with open(path.join(this_directory, "LICENSE")) as f:
    license_text = f.read()


setup(
    name="pypulsepal",
    version="0.0.1",
    description="pypulsepal",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    packages=find_packages(),
    url="https://github.com/larsrollik/pypulsepal",
    author="Lars B. Rollik",
    author_email="L.B.Rollik@protonmail.com",
    license=license_text,
    install_requires=[
        "tqdm",
        "pybpod-api",  # for ArCOM
    ],
    extras_require={
        "dev": [
            "black",
            "pytest-cov",
            "pytest",
            "gitpython",
            "coverage>=5.0.3",
            "bump2version",
            "pre-commit",
            "flake8",
        ],
    },
    zip_safe=False,
    include_package_data=True,
)



import logging

from pybpodapi.com.arcom import ArCOM
from tqdm import tqdm

from pypulsepal.definitions import CHANNEL_PARAM_DEFAULTS
from pypulsepal.definitions import CUSTOM_PULSE_TRAIN_OPCODES
from pypulsepal.definitions import PARAM_DTYPE_MODEL_1
from pypulsepal.definitions import PARAM_DTYPE_MODEL_2
from pypulsepal.definitions import PARAM_SCALING
from pypulsepal.definitions import PULSEPAL_CYCLE_FREQUENCY
from pypulsepal.definitions import ReceiveMessageHeader
from pypulsepal.definitions import resolve_param_name_code_pair
from pypulsepal.definitions import resolve_trigger_name_code_pair
from pypulsepal.definitions import SendMessageHeader
from pypulsepal.definitions import TRIGGER_PARAM_DEFAULTS
from pypulsepal.utils import encode_message
from pypulsepal.utils import volts_to_bytes


class PulsePalError(Exception):
    """Convenience error object for PulsePal"""

    pass


class PulsePal:
    """"""

    # communication
    _arcom = None
    serial_port = None
    baudrate = 115200

    # hardware attributes
    firmware_version = None
    model = None
    dac_bitMax = None
    cycle_frequency = 20000
    nr_output_channels = 4
    nr_trigger_channels = 2
    opcode = 213
    param_dtype_lookup = None

    def __init__(
        self,
        serial_port=None,
        baudrate=115200,
        cycle_frequency=PULSEPAL_CYCLE_FREQUENCY,
        nr_output_channels=4,
        nr_trigger_channels=2,
        opcode=213,
        **kwargs,
    ):
        """

        :param serial_port:
        :param baudrate:
        :param cycle_frequency:
        :param nr_output_channels:
        :param nr_trigger_channels:
        :param kwargs:
        """
        super(PulsePal, self).__init__()

        # Generate class attributes according to default channel parameters
        for param, default_value in CHANNEL_PARAM_DEFAULTS.items():
            setattr(self, param, [default_value] * self.nr_output_channels)

        for param, default_value in TRIGGER_PARAM_DEFAULTS.items():
            setattr(self, param, [default_value] * self.nr_trigger_channels)

        # Args
        self.serial_port = serial_port
        self.baudrate = baudrate
        self.cycle_frequency = cycle_frequency
        self.nr_output_channels = nr_output_channels
        self.nr_trigger_channels = nr_trigger_channels
        self.opcode = opcode

        # Convenience updates for debug inputs
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

        self.connect(serial_port=serial_port, baudrate=baudrate)

    @property
    def encoded_opcode(self):
        return encode_message(self.opcode, encoding="uint8")

    @encoded_opcode.setter
    def encoded_opcode(self, value=None):
        self.opcode = value

    def _clear_read_queue(self):
        """Clears leftover items from serial read queue"""
        return self._arcom.serial_object.read(self._arcom.serial_object.inWaiting())

    def _read_confirmation(self):
        """Returns True for successful receipt of previous message"""
        return self._arcom.read_uint8() == 1

    def _pulsepal_handshake(self):
        """Confirm connectivity with hardware.

        :return: handshake success bool
        """
        self._arcom.serial_object.write(
            self.encoded_opcode + str.encode(SendMessageHeader.HANDSHAKE)
        )
        handshake, firmware_version = self._arcom.read_char(), self._arcom.read_uint8()
        self._clear_read_queue()

        handshake_ok = handshake == ReceiveMessageHeader.HANDSHAKE_OK
        if handshake_ok:
            self.firmware_version = firmware_version
            if firmware_version < 20:
                self.model = 1
                self.dac_bitMax = 255
                self.param_dtype_lookup = PARAM_DTYPE_MODEL_1
            else:
                self.model = 2
                self.dac_bitMax = 65535
                self.param_dtype_lookup = PARAM_DTYPE_MODEL_2

        return True if handshake_ok else False

    def connect(self, serial_port, baudrate=115200, timeout=1):
        """Connect (& handshake) with hardware

        :param serial_port:
        :param baudrate:
        :param timeout:
        :return:
        """
        self._arcom = ArCOM().open(
            serial_port=serial_port, baudrate=baudrate, timeout=timeout
        )
        handshake_ok = self._pulsepal_handshake()
        if not handshake_ok:
            raise PulsePalError(
                f"Could not connect PulsePal at '{serial_port}' with baudrate {baudrate}"
            )
        return self

    def _pulsepal_set_display(self, message="--> Py"):
        """"""
        message_ascii = [ord(s) for s in message]
        self._arcom.write_array(
            *[
                self.encoded_opcode,
                encode_message(SendMessageHeader.DISPLAY, encoding="uint8"),
                *message_ascii,
            ]
        )

    def _update_param(self, channel, param_name, param_value):
        attr = getattr(self, param_name)
        attr[channel] = param_value
        setattr(self, param_name, attr)

    def program_one_param(self, channel=None, param_name=None, param_value=None):
        """Program one channel parameter (one parameter on one channel)."""
        param_name, param_code = resolve_param_name_code_pair(
            param_name_or_code=param_name
        )
        param_dtype, param_scaling = self.param_dtype_lookup.get(
            param_name
        ), PARAM_SCALING.get(param_name)

        logging.debug(f"Param value before voltage-to-bit correction: {param_value}")
        if "volt" in param_name.lower():
            param_value = volts_to_bytes(volt=param_value, dac_bitMax=self.dac_bitMax)

        logging.debug(f"Param value before time scaling correction: {param_value}")
        param_value = param_value * param_scaling

        # Send
        message = [
            self.encoded_opcode,
            encode_message(SendMessageHeader.PROGRAM_ONE, encoding="uint8"),
            encode_message(param_code, encoding="uint8"),
            encode_message(channel + 1, encoding="uint8"),
            encode_message(param_value, encoding=param_dtype),
        ]
        self._arcom.write_array(b"".join(message))

        write_ok = self._read_confirmation()
        if write_ok:
            self._update_param(
                channel=channel, param_name=param_name, param_value=param_value
            )

        return write_ok

    def program_trigger_channel(self, trigger_channel=None, trigger_mode=None):
        """"""
        _, trigger_mode_value = resolve_trigger_name_code_pair(
            trigger_name_or_code=trigger_mode
        )
        write_ok = self.program_one_param(
            channel=trigger_channel,
            param_name="triggerMode",
            param_value=trigger_mode_value,
        )
        return write_ok

    def upload_all(self):
        """Set each channel and trigger param on each output channel and trigger channel, respectively."""
        # Channel parameters
        for param_name in tqdm(CHANNEL_PARAM_DEFAULTS):
            for channel in range(self.nr_output_channels):

                param_value = getattr(self, param_name)[channel]
                success = self.program_one_param(
                    channel=channel, param_name=param_name, param_value=param_value
                )
                logging.debug(param_name, channel, param_value, f"ok: {success}")
                if not success:
                    raise ValueError

        # Trigger modes
        param_name = "triggerMode"
        for trigger_channel in tqdm(range(self.nr_trigger_channels)):
            param_value = getattr(self, param_name)[trigger_channel]
            success = self.program_one_param(
                channel=trigger_channel, param_name=param_name, param_value=param_value
            )
            logging.debug(param_name, trigger_channel, param_value, f"ok: {success}")
            if not success:
                raise ValueError

    def set_resting_voltage(self, channel=None, voltage=None):
        """Convenience function to set restingVoltage parameter on one channel.

        :param channel:
        :param voltage:
        :return:
        """
        write_ok = self.program_one_param(
            channel=channel,
            param_name="restingVoltage",
            param_value=voltage,
        )
        return write_ok

    def upload_custom_pulse_train(
        self, pulse_train_id=None, pulse_times=None, pulse_voltages=None
    ):
        """"""
        assert pulse_train_id in [0, 1]
        assert len(pulse_times) == len(pulse_voltages)

        scaled_pulse_times = []
        scaled_pulse_voltages = []
        for pulse_time, pulse_voltage in zip(pulse_times, pulse_voltages):
            scaled_pulse_times.append(pulse_time * self.cycle_frequency)
            scaled_pulse_voltages.append(
                volts_to_bytes(volt=pulse_voltage, dac_bitMax=self.dac_bitMax)
            )

        message = [
            self.encoded_opcode,
            encode_message(
                CUSTOM_PULSE_TRAIN_OPCODES.get(pulse_train_id), encoding="uint8"
            ),
            # if model=1, additional 0 int here,
            encode_message(len(scaled_pulse_times), encoding="uint32"),
            encode_message(scaled_pulse_times, encoding="uint32"),
            encode_message(
                scaled_pulse_voltages,
                encoding=self.param_dtype_lookup.get("phase1Voltage"),
            ),
        ]
        self._arcom.write_array(b"".join(message))
        return self._read_confirmation()

    def upload_custom_waveform(
        self, pulse_train_id=None, pulse_width=None, pulse_voltages=None
    ):
        """"""
        assert pulse_train_id in [0, 1]

        scaled_pulse_times = []
        scaled_pulse_voltages = []
        for pulse_index, pulse_voltage in enumerate(pulse_voltages):
            pulse_time = pulse_index * pulse_width * self.cycle_frequency
            scaled_pulse_times.append(pulse_time)
            scaled_pulse_voltages.append(
                volts_to_bytes(volt=pulse_voltage, dac_bitMax=self.dac_bitMax)
            )

        message = [
            self.encoded_opcode,
            encode_message(
                CUSTOM_PULSE_TRAIN_OPCODES.get(pulse_train_id), encoding="uint8"
            ),
            # if model=1, additional 0 int here,
            encode_message(len(scaled_pulse_times), encoding="uint32"),
            encode_message(scaled_pulse_times, encoding="uint32"),
            encode_message(
                scaled_pulse_voltages,
                encoding=self.param_dtype_lookup.get("phase1Voltage"),
            ),
        ]
        self._arcom.write_array(b"".join(message))
        return self._read_confirmation()

    def set_continuous(self, channel=None, state=None):
        """"""
        message = [
            self.encoded_opcode,
            encode_message(SendMessageHeader.CONTINUOUS, encoding="uint8"),
            encode_message(channel, encoding="uint8"),
            encode_message(state, encoding="uint8"),
        ]
        self._arcom.write_array(b"".join(message))
        return self._read_confirmation()

    def trigger_selected_channels(
        self, channel_1=False, channel_2=False, channel_3=False, channel_4=False
    ):
        """Trigger specific channels

        :param channel_1:
        :param channel_2:
        :param channel_3:
        :param channel_4:
        :return:
        """
        combination_byte = (
            (1 * channel_1) + (2 * channel_2) + (4 * channel_3) + (8 * channel_4)
        )
        message = [
            self.encoded_opcode,
            encode_message(SendMessageHeader.SOFT_TRIGGER, encoding="uint8"),
            encode_message(combination_byte, encoding="uint8"),
        ]
        self._arcom.write_array(b"".join(message))
        return self._read_confirmation()

    def trigger_all_channels(self):
        return self.trigger_channels(
            channel_1=True, channel_2=True, channel_3=True, channel_4=True
        )

    def stop_all_outputs(self):
        """"""
        message = [
            self.encoded_opcode,
            encode_message(SendMessageHeader.ABORT_ALL, encoding="uint8"),
        ]
        self._arcom.write_array(b"".join(message))
        return self._read_confirmation()

    def save_settings(self):
        """"""
        message = [
            self.encoded_opcode,
            encode_message(SendMessageHeader.DISCONNECT, encoding="uint8"),
        ]
        self._arcom.write_array(b"".join(message))
        return self._read_confirmation()  # fixme: returns False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save_settings()
        self._arcom.close()

    def __del__(self):
        self.save_settings()
        self._arcom.close()
